#!moo verb @listgag listgag --on $player

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is used as a player command to list the objects in the player's gag lists. If the gag lists are not empty, a
list of the names of the players in the `gaglist` property on the player, and the objects in the `object_gaglist`
property are produced.

The verb uses a test based on the `context.parser` object, to determine if it is being called by another verb, or
invoked by a player as a command. If `context.parser` returns `None` (and hence is not `True`), then the verb is being
invoked as a player command. When False, the verb is being invoked by another verb, and therefore returns JSON.

TODO: In addition, the verb checks through the gag lists of all the player's in the database, to see if the player who
TODO: invoked the command is listed in anyone else's gag list. If this is the case, a list of the people doing the
TODO: gagging is printed.
"""

from moo.core import context

player = context.player

if not context.parser:
    print("Gagged players:")
    if player.gaglist:
        for p in player.gaglist:
            print(f"  {p.name}")
    else:
        print("  None")

    print("Gagged objects:")
    if player.object_gaglist:
        for o in player.object_gaglist:
            print(f"  {o.name}")
    else:
        print("  None")
else:
    return {"gagged_players": player.gaglist, "gagged_objects": player.object_gaglist}



#!moo verb page_absent_msg page_origin_msg page_echo_msg whereis_location_msg who_location_msg --on $player

# pylint: disable=undefined-variable,return-outside-function

"""
These verbs return a pronoun substituted version of the corresponding properties stored on the player object.
They are used by the `page` verb, and the `whereis` and `who` commands.
"""

if verb_name == "page_absent_msg":
    return _.string_utils.pronoun_sub(this.get_property("page_absent_msg"))
elif verb_name == "page_origin_msg":
    return _.string_utils.pronoun_sub(this.get_property("page_origin_msg"))
elif verb_name == "page_echo_msg":
    return _.string_utils.pronoun_sub(this.get_property("page_echo_msg"))
elif verb_name == "whereis_location_msg":
    return _.string_utils.pronoun_sub(this.get_property("whereis_location_msg"), this)
elif verb_name == "who_location_msg":
    return _.string_utils.pronoun_sub(this.get_property("who_location_msg"))
else:
    raise ValueError(f"Unknown verb name: {verb_name}")



# -*- coding: utf-8 -*-
"""
Useful global utilities.
"""


def _make_access(instance, rule, permission_id, group):
    from .models.acl import Access
    return Access(
        object=instance if instance.kind == "object" else None,
        verb=instance if instance.kind == "verb" else None,
        property=instance if instance.kind == "property" else None,
        rule=rule,
        permission_id=permission_id,
        type="group",
        group=group,
    )


def apply_default_permissions(instance):
    """
    Apply default permissions to a newly created Object, Verb, or Property.

    This mirrors the logic of the ``set_default_permissions`` verb natively,
    avoiding verb-lookup and execution overhead. The verb is considered stable
    for the lifetime of a running service; changes to it require a matching
    update here and a service restart.
    """
    from .models.acl import Access, _get_permission_id
    anything_id = _get_permission_id("anything")
    perm_for_everyone = _get_permission_id("execute" if instance.kind == "verb" else "read")
    Access.objects.bulk_create([
        _make_access(instance, "allow", anything_id, "wizards"),
        _make_access(instance, "allow", anything_id, "owners"),
        _make_access(instance, "allow", perm_for_everyone, "everyone"),
    ])


def apply_default_permissions_bulk(instances):
    """
    Apply default permissions to many newly created instances in a single query.
    Equivalent to calling :func:`apply_default_permissions` for each instance,
    but emits only one ``INSERT`` for all ACL records combined.
    """
    from .models.acl import Access, _get_permission_id
    anything_id = _get_permission_id("anything")
    read_id = _get_permission_id("read")
    execute_id = _get_permission_id("execute")
    records = []
    for instance in instances:
        perm_for_everyone = execute_id if instance.kind == "verb" else read_id
        records.extend([
            _make_access(instance, "allow", anything_id, "wizards"),
            _make_access(instance, "allow", anything_id, "owners"),
            _make_access(instance, "allow", perm_for_everyone, "everyone"),
        ])
    if records:
        Access.objects.bulk_create(records)

def expand_wildcard(name):
    if "*" not in name:
        return [name]
    prefix, suffix = name.split("*", 1)
    result = [prefix]
    for i in range(len(suffix)):
        result.append(prefix + suffix[:i+1])
    return result



# Generated manually for performance improvements
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0022_delete_accessibleobject_delete_accessibleproperty_and_more"),
    ]

    operations = [
        migrations.AddIndex(
            model_name="property",
            index=models.Index(fields=["origin", "name"], name="property_origin_name_idx"),
        ),
        migrations.AddIndex(
            model_name="access",
            index=models.Index(fields=["object", "permission"], name="access_object_permission_idx"),
        ),
        migrations.AddIndex(
            model_name="access",
            index=models.Index(fields=["verb", "permission"], name="access_verb_permission_idx"),
        ),
        migrations.AddIndex(
            model_name="access",
            index=models.Index(fields=["property", "permission"], name="access_property_permission_idx"),
        ),
        migrations.AddIndex(
            model_name="player",
            index=models.Index(fields=["avatar", "wizard"], name="player_avatar_wizard_idx"),
        ),
    ]



#!moo verb confunc --on $room

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is called by the LambdaMOO server when a player connects in a room.

The action coded into the $room class simply shows the player what the room looks like (using the `look_self` verb of
the room) and tells everyone else in the room that the player has connected.
"""

from moo.core import context

this.look_self()
this.announce(f"{context.player} has connected.")



import pytest

from .. import code, create
from ..exceptions import QuotaError
from ..models import Object


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_create_object(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer):
        obj = create("widget")
    assert obj.name == "widget"
    assert obj.pk > 0


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_create_object_with_quota(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    t_wizard.set_property("ownership_quota", 1)
    with code.ContextManager(t_wizard, _writer):
        obj = create("widget")
        with pytest.raises(QuotaError):
            create("widget 2")
    quota = t_wizard.get_property("ownership_quota")
    assert quota == 0
    assert obj.name == "widget"
    assert obj.pk > 0


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_delete_object_with_quota(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    t_wizard.set_property("ownership_quota", 2)
    with code.ContextManager(t_wizard, _writer):
        obj = create("widget")
    quota = t_wizard.get_property("ownership_quota")
    assert quota == 1
    with code.ContextManager(t_wizard, _writer):
        obj.delete()
    quota = t_wizard.get_property("ownership_quota")
    assert quota == 2



#!moo verb recycle --on $root_class

# pylint: disable=return-outside-function,undefined-variable

"""
This verb contains no code for the $root class. It is called by the obj.delete() primitive just before an object is
removed. This is useful to make sure that deleting objects does not leave the database in a strange state. For
example, the $exit class uses the `recycle` verb to remove the exit from the the entrance and exit lists of its
destination and source rooms.

By default, the `recycle` verb on $root does nothing.
"""

pass



import logging
from time import time

from moo import bootstrap
from moo.core import code, create, lookup
from moo.core.models import Player, User

log = logging.getLogger(__name__)

repo = bootstrap.initialize_dataset("default")
wizard = lookup("Wizard")
user = User.objects.create(username="wizard")
Player.objects.create(user=user, avatar=wizard, wizard=True)

containers = lookup("container class")
containers.name = "Generic Container"
containers.save()

with code.ContextManager(wizard, log.info):
    sys = lookup(1)

    sys.set_property("container", containers)
    containers.set_property("open_key", None)
    containers.set_property("open", False, inherit_owner=True)
    containers.set_property("opaque", False, inherit_owner=True)

    root = create("Root Class", location=None)
    sys.set_property("root_class", root)
    root.add_verb("accept", code="return True")
    root.set_property("description", "")
    root.set_property("key", None)

    rooms = create("Generic Room", parents=[root], location=None)
    sys.set_property("room", rooms)
    rooms.set_property("blessed_object", None)
    rooms.set_property("blessed_task_id", None)
    rooms.set_property("description", "There's not much to see here.")
    rooms.set_property("exits", [])
    rooms.set_property("victim_ejection_message", "You have been ejected from %s by %s.")
    rooms.set_property("ejection_message", "You have ejected %s.")
    rooms.set_property("room_ejection_message", "%s has been ejected by %s.")
    rooms.set_property("residents", [])
    rooms.set_property("free_entry", True)
    rooms.set_property("entrances", [])
    rooms.set_property("dark", False)
    rooms.set_property("content_list_type", 3)

    mail_room = create("Mail Distribution Center", parents=[rooms], location=None)

    player = create("Generic Player", parents=[root], location=None)
    sys.set_property("player", player)
    player.set_property("last_connect_time", time())
    player.set_property("ownership_quota", 1000)
    player.set_property("ps", "they", inherit_owner=True)
    player.set_property("po", "them", inherit_owner=True)
    player.set_property("pp", "their", inherit_owner=True)
    player.set_property("pr", "themselves", inherit_owner=True)
    player.set_property("pq", "theirs", inherit_owner=True)
    player.set_property("psc", "They", inherit_owner=True)
    player.set_property("poc", "Them", inherit_owner=True)
    player.set_property("ppc", "Their", inherit_owner=True)
    player.set_property("prc", "Themselves", inherit_owner=True)
    player.set_property("pqc", "Theirs", inherit_owner=True)
    player.set_property("home", None)
    player.set_property("gaglist", [])
    player.set_property("paranoid", False)
    player.set_property("responsible", [])
    player.set_property("lines", 10)
    player.set_property("object_gaglist", [])
    player.set_property("page_absent_msg", "%n is not currently logged in.")
    player.set_property("page_origin_msg", "You sense that %n is looking for you in %x(location).")
    player.set_property("page_echo_msg", "Your message has been sent.")
    player.set_property("whereis_location_msg", "%N is in %x(location).")
    player.set_property("who_location_msg", "%x(location)")
    player.set_property("victim_ejection_msg", "You have been ejected from %s by %s.")

    programmers = create("Generic Programmer", parents=[player], location=None)
    sys.set_property("programmer", programmers)
    wizards = create("Generic Wizard", parents=[programmers], location=None)
    sys.set_property("wizard", wizards)
    wizard.parents.add(wizards)

    thing = create("Generic Thing", parents=[root], location=None)
    sys.set_property("thing", thing)
    thing.set_property("otake_succeeded_msg", "{actor} picks up {subject}.")
    thing.set_property("otake_failed_msg", "")
    thing.set_property("take_succeeded_msg", "You take {subject}.")
    thing.set_property("take_failed_msg", "You can't pick that up.")
    thing.set_property("odrop_succeeded_msg", "{actor} drops {subject}.")
    thing.set_property("odrop_failed_msg", "{actor} tries to drop {subject} but fails!")
    thing.set_property("drop_succeeded_msg", "You drop {subject}.")
    thing.set_property("drop_failed_msg", "You can't seem to drop {subject} here.")
    containers.parents.add(thing)

    exits = create("Generic Exit", parents=[root], location=None)
    sys.set_property("exit", exits)
    exits.set_property("source", None)
    exits.set_property("dest", None)
    exits.set_property("nogo_msg", "You can't go that way.")
    exits.set_property("onogo_msg", "{actor} can't go that way.")
    exits.set_property("arrive_msg", "You arrive at {subject}.")
    exits.set_property("oarrive_msg", "{actor} arrives at {subject}.")
    exits.set_property("oleave_msg", "{actor} leaves {subject}.")
    exits.set_property("leave_msg", "You leave {subject}.")

    notes = create("Generic Note", parents=[thing], location=None)
    sys.set_property("note", notes)
    letters = create("Generic Letter", parents=[notes], location=None)
    sys.set_property("letter", letters)

    lock_utils = create("Lock Utilities", location=None)
    sys.set_property("lock_utils", lock_utils)

    string_utils = create("String Utilities", location=None)
    sys.set_property("string_utils", string_utils)

    gender_utils = create("Gender Utilities", location=None)
    sys.set_property("gender_utils", gender_utils)
    gender_utils.set_property("pronouns", [
        "ps", "po", "pp", "pr", "pq", "psc", "poc", "ppc", "prc", "pqc"
    ])
    gender_utils.set_property("genders", ["neuter", "male", "female", "plural"])
    gender_utils.set_property("ps", ["it", "he", "she", "they"])
    gender_utils.set_property("po", ["it", "him", "her", "them"])
    gender_utils.set_property("pp", ["its", "his", "her", "their"])
    gender_utils.set_property("pr", ["itself", "himself", "herself", "themselves"])
    gender_utils.set_property("pq", ["its", "his", "hers", "theirs"])
    gender_utils.set_property("psc", ["It", "He", "She", "They"])
    gender_utils.set_property("poc", ["It", "Him", "Her", "Them"])
    gender_utils.set_property("ppc", ["Its", "His", "Hers", "Their"])
    gender_utils.set_property("prc", ["Itself", "Himself", "Herself", "Themselves"])
    gender_utils.set_property("pqc", ["Its", "His", "Hers", "Theirs"])

    lab = create("The Laboratory", parents=[rooms], location=None)
    lab.set_property(
        "description",
        """A cavernous laboratory filled with gadgetry of every kind,
    this seems like a dumping ground for every piece of dusty forgotten
    equipment a mad scientist might require.""",
    )
    sys.set_property("player_start", lab)

    wizard.location = lab
    wizard.save()

    new_player = create(name="Player", unique_name=True, location=lab)
    p = Player.objects.create()
    p.avatar = new_player
    p.save()

    new_player.parents.add(player)
    new_player.owner = new_player
    new_player.save()

    root.get_verb("accept").delete()

    containers.get_verb("accept").delete()
    bootstrap.load_verbs(repo, "moo.bootstrap.default_verbs")
    sys.gender_utils.set(player, "plural")



# -*- coding: utf-8 -*-
"""
Task model.
"""

from django.db import models


class Task(models.Model):
    user = models.ForeignKey("Object", related_name="tasks", on_delete=models.CASCADE)
    origin = models.ForeignKey("Object", related_name="+", on_delete=models.CASCADE)
    verb_name = models.CharField(max_length=255)
    args = models.TextField()
    kwargs = models.TextField()
    created = models.DateTimeField()
    delay = models.IntegerField()
    killed = models.BooleanField(default=False)
    error = models.CharField(max_length=255, blank=True, null=True)
    trace = models.TextField(blank=True, null=True)



# -*- coding: utf-8 -*-
"""
Development support resources for MOO programs
"""

import contextvars
import functools
import logging
import warnings

from django.conf import settings
from RestrictedPython import compile_restricted, compile_restricted_function
from RestrictedPython.Guards import (guarded_iter_unpack_sequence,
                                     guarded_unpack_sequence, safe_builtins)

log = logging.getLogger(__name__)


def interpret(source, name, *args, runtype="exec", **kwargs):
    from . import context

    globals = get_default_globals()  # pylint: disable=redefined-builtin
    globals.update(get_restricted_environment(name, context.writer))
    if runtype == "exec":
        return r_exec(source, {}, globals, *args, **kwargs)
    else:
        return r_eval(source, {}, globals, *args, **kwargs)


@functools.lru_cache(maxsize=512)
def _cached_compile(body, filename):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=SyntaxWarning)
        return compile_restricted_function(
            p="this=None, passthrough=None, _=None, *args, **kwargs",
            body=body,
            name="verb",
            filename=filename,
        )


def compile_verb_code(body, filename):
    """
    Take a given piece of verb code and wrap it in a function.
    """
    return _cached_compile(body, filename)


def r_eval(src, locals, globals, *args, filename="<string>", **kwargs):  # pylint: disable=redefined-builtin
    code = compile_verb_code(src, filename)
    return do_eval(code, locals, globals, *args, filename=filename, runtype="eval", **kwargs)


def r_exec(src, locals, globals, *args, filename="<string>", **kwargs):  # pylint: disable=redefined-builtin
    code = compile_verb_code(src, filename)
    return do_eval(code, locals, globals, *args, filename=filename, runtype="exec", **kwargs)


def do_eval(
    code, locals, globals, *args, filename="<string>", runtype="eval", **kwargs
):  # pylint: disable=redefined-builtin
    """
    Execute an expression in the provided environment.
    """
    if isinstance(code, str):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=SyntaxWarning)
            code = compile_restricted(code, filename, runtype)

        value = eval(code, globals, locals)  # pylint: disable=eval-used
    else:
        exec(code.code, globals, locals)  # pylint: disable=exec-used
        compiled_function = locals["verb"]
        value = compiled_function(*args, **kwargs)
    return value


def get_default_globals():
    return {"__name__": "__main__", "__package__": None, "__doc__": None}


def get_restricted_environment(name, writer):
    """
    Construct an environment dictionary.
    """

    class _print_:
        def _call_print(self, s):
            writer(s)

    class _write_:
        def __init__(self, obj):
            object.__setattr__(self, "obj", obj)

        def __setattr__(self, name, value):
            """
            Private attribute protection using is_frame_access_allowed()
            """
            set_protected_attribute(self.obj, name, value)  # pylint: disable=no-member

        def __setitem__(self, key, value):
            """
            Passthrough property access.
            """
            self.obj[key] = value  # pylint: disable=no-member

    def restricted_import(name, gdict, ldict, fromlist, level=-1):
        """
        Used to drastically limit the importable modules.
        """
        if name in settings.ALLOWED_MODULES:
            return __builtins__["__import__"](name, gdict, ldict, fromlist, level)
        raise ImportError("Restricted: %s" % name)

    def get_protected_attribute(obj, name, g=getattr):
        if name.startswith("_"):
            raise AttributeError(name)
        return g(obj, name)

    def set_protected_attribute(obj, name, value, s=setattr):
        if name.startswith("_"):
            raise AttributeError(name)
        return s(obj, name, value)

    def inplace_var_modification(operator, a, b):
        if operator == "+=":
            return a + b
        raise NotImplementedError("In-place modification with %s not supported." % operator)

    safe_builtins["__import__"] = restricted_import

    for n in settings.ALLOWED_BUILTINS:
        safe_builtins[n] = __builtins__[n]
    env = dict(
        _apply_=lambda f, *a, **kw: f(*a, **kw),
        _print_=lambda x: _print_(),
        _print=_print_(),
        _write_=_write_,
        _getattr_=get_protected_attribute,
        _getitem_=lambda obj, key: obj[key],
        _getiter_=iter,
        _inplacevar_=inplace_var_modification,
        _unpack_sequence_=guarded_unpack_sequence,
        _iter_unpack_sequence_=guarded_iter_unpack_sequence,
        __import__=restricted_import,
        __builtins__=safe_builtins,
        __metaclass__=type,
        __name__=name,
        __package__=None,
        __doc__=None,
        verb_name=name
    )

    return env


_active_caller = contextvars.ContextVar("active_caller", default=None)
_active_player = contextvars.ContextVar("active_player", default=None)
_active_writer = contextvars.ContextVar("active_writer", default=None)
_active_parser = contextvars.ContextVar("active_parser", default=None)
_active_task_id = contextvars.ContextVar("active_task_id", default=None)
_verb_lookup_cache = contextvars.ContextVar("verb_lookup_cache", default=None)
_prop_lookup_cache = contextvars.ContextVar("prop_lookup_cache", default=None)
# A sentinel object (not a mutable default like []) so we can reliably detect whether
# a ContextManager is active: _active_caller_stack.get() is _UNSET means no session.
# Using a mutable list as the default would cause shared-state bugs — any code that
# appended to it outside a context would mutate the single default object permanently.
_UNSET = object()
_active_caller_stack = contextvars.ContextVar("active_caller_stack", default=_UNSET)
_perm_cache = contextvars.ContextVar("perm_cache", default=None)

class ContextManager:
    """
    The ContextManager class is what holds critical per-execution information such as
    the active user and writer.  It uses contextvars to maintain this information across
    asynchronous calls.

    This contextmanager should really only be used once per top-level request, such as
    when a user invokes a verb or issues a command in the console. Nested uses of this
    contextmanager are supported for unit testing purposes, since eager Celery execution
    means that verb invocations within verbs happen synchronously.
    """
    @classmethod
    def get(cls, name):
        if name == "caller":
            return _active_caller.get()
        if name == "player":
            return _active_player.get()
        if name == "writer":
            return _active_writer.get()
        if name == "parser":
            return _active_parser.get()
        if name == "task_id":
            return _active_task_id.get()
        if name == "caller_stack":
            stack = _active_caller_stack.get()
            # Return an empty list outside a session rather than exposing the sentinel.
            return [] if stack is _UNSET else stack
        raise NotImplementedError(f"Unknown ContextManager variable: {name}")

    @classmethod
    def get_perm_cache(cls) -> dict | None:
        return _perm_cache.get()

    @classmethod
    def get_verb_lookup_cache(cls) -> dict | None:
        return _verb_lookup_cache.get()

    @classmethod
    def get_prop_lookup_cache(cls) -> dict | None:
        return _prop_lookup_cache.get()

    @classmethod
    def is_active(cls):
        # True only when __enter__ has been called and __exit__ has not yet run.
        # Verb.__call__ uses this to skip stack tracking for utility verb calls
        # that happen outside any user session (e.g. apply_default_permissions).
        return _active_caller_stack.get() is not _UNSET

    @classmethod
    def override_caller(cls, caller, this=None, verb_name=None, origin=None, player=None):
        attributes = {
            "this": this,
            "verb_name": verb_name,
            "caller": caller,
            "origin": origin,
            "player": player,
            "previous_caller": _active_caller.get(),
        }
        _active_caller.set(caller)
        caller_stack = _active_caller_stack.get()
        caller_stack.append(attributes)
        _active_caller_stack.set(caller_stack)

    @classmethod
    def pop_caller(cls):
        caller_stack = _active_caller_stack.get()
        if not caller_stack:
            raise RuntimeError("Caller stack is empty.")
        frame = caller_stack.pop()
        _active_caller_stack.set(caller_stack)
        _active_caller.set(frame["previous_caller"])

    def __init__(self, caller, writer, task_id=None):
        self.caller = caller
        self.caller_token = None
        self.player = self.caller
        self.player_token = None
        self.writer = writer
        self.writer_token = None
        self.parser = None
        self.parser_token = None
        self.task_id = task_id
        self.task_id_token = None
        # A fresh list per instance so each session has its own isolated stack.
        self.active_caller_stack = []
        self.active_caller_stack_token = None
        # A fresh dict per instance for per-session permission caching.
        self.perm_cache = {}
        self.perm_cache_token = None
        # Per-session verb and property lookup caches keyed by (object_pk, name, ...).
        self.verb_lookup_cache = {}
        self.verb_lookup_cache_token = None
        self.prop_lookup_cache = {}
        self.prop_lookup_cache_token = None

    def set_parser(self, parser):
        self.parser = parser
        self.parser_token = _active_parser.set(self.parser)

    def __enter__(self):
        self.caller_token = _active_caller.set(self.caller)
        self.player_token = _active_player.set(self.player)
        self.writer_token = _active_writer.set(self.writer)
        self.task_id_token = _active_task_id.set(self.task_id)
        self.parser_token = _active_parser.set(self.parser)
        # Replacing _UNSET with a real list marks the session as active (is_active() → True)
        # and gives override_caller a mutable list to append to.
        self.active_caller_stack_token = _active_caller_stack.set(self.active_caller_stack)
        self.perm_cache_token = _perm_cache.set(self.perm_cache)
        self.verb_lookup_cache_token = _verb_lookup_cache.set(self.verb_lookup_cache)
        self.prop_lookup_cache_token = _prop_lookup_cache.set(self.prop_lookup_cache)
        return self

    def __exit__(self, cls, value, traceback):
        if self.caller_token:
            _active_caller.reset(self.caller_token)
        if self.player_token:
            _active_player.reset(self.player_token)
        if self.writer_token:
            _active_writer.reset(self.writer_token)
        if self.parser_token:
            _active_parser.reset(self.parser_token)
        if self.task_id_token:
            _active_task_id.reset(self.task_id_token)
        if self.active_caller_stack_token:
            # reset() restores to whatever the var held before __enter__'s set() call,
            # regardless of how many times override_caller/pop_caller mutated it since.
            _active_caller_stack.reset(self.active_caller_stack_token)
        if self.perm_cache_token:
            _perm_cache.reset(self.perm_cache_token)
        if self.verb_lookup_cache_token:
            _verb_lookup_cache.reset(self.verb_lookup_cache_token)
        if self.prop_lookup_cache_token:
            _prop_lookup_cache.reset(self.prop_lookup_cache_token)



"""
Management command to rebuild the AncestorCache table from scratch.

Use this after bulk Relationship changes, data migrations, or to recover from
a signal failure that left the cache inconsistent.
"""
from django.core.management.base import BaseCommand
from django_cte import with_cte

from moo.core.models.object import (
    AncestorCache,
    Object,
    _make_ancestors_cte,
)


class Command(BaseCommand):
    help = "Rebuild the AncestorCache table from scratch using the current Relationship data."

    def handle(self, *args, **options):
        self.stdout.write("Clearing existing AncestorCache rows...")
        deleted, _ = AncestorCache.objects.all().delete()
        self.stdout.write(f"  Deleted {deleted} rows.")

        all_pks = list(Object.objects.values_list("id", flat=True))
        self.stdout.write(f"Rebuilding cache for {len(all_pks)} objects...")

        rows = []
        processed = 0
        for pk in all_pks:
            ancestors_cte = _make_ancestors_cte(pk)
            seen = {}
            for row in with_cte(
                ancestors_cte,
                select=ancestors_cte.join(Object, id=ancestors_cte.col.object_id)
                .annotate(depth=ancestors_cte.col.depth, path_weight=ancestors_cte.col.path_weight)
                .order_by("depth", "-path_weight")
            ).values("id", "depth", "path_weight"):
                if row["id"] not in seen:
                    seen[row["id"]] = row
            for row in seen.values():
                rows.append(AncestorCache(
                    descendant_id=pk,
                    ancestor_id=row["id"],
                    depth=row["depth"],
                    path_weight=row["path_weight"],
                ))
            processed += 1
            if len(rows) >= 500:
                AncestorCache.objects.bulk_create(rows, ignore_conflicts=True)
                rows = []

        if rows:
            AncestorCache.objects.bulk_create(rows, ignore_conflicts=True)

        total = AncestorCache.objects.count()
        self.stdout.write(
            self.style.SUCCESS(
                f"Done. Processed {processed} objects, created {total} AncestorCache rows."
            )
        )



import pytest

from moo.core import code, create, lookup, parse
from moo.core.models import Object


def setup_container(t_wizard: Object):
    containers = lookup("Generic Container")
    box = create("wooden box", parents=[containers], location=t_wizard.location)
    return box


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_is_open_default(t_init: Object, t_wizard: Object):
    """Container starts closed by default."""
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer):
        box = setup_container(t_wizard)
        assert not box.is_open()


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_open_container(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer) as ctx:
        box = setup_container(t_wizard)
        parse.interpret(ctx, "open wooden box")
        box.refresh_from_db()
        assert printed == ["You open the container."]
        assert box.is_open()


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_open_already_open(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer) as ctx:
        box = setup_container(t_wizard)
        parse.interpret(ctx, "open wooden box")
        printed.clear()
        parse.interpret(ctx, "open wooden box")
        assert printed == ["Container is already open."]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_close_container(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer) as ctx:
        box = setup_container(t_wizard)
        assert not box.is_open()
        parse.interpret(ctx, "open wooden box")
        assert box.is_open()
        printed.clear()
        parse.interpret(ctx, "close wooden box")
        box.refresh_from_db()
        assert not box.is_open()
        assert printed == ["You close the container."]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_close_already_closed(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer) as ctx:
        box = setup_container(t_wizard)
        parse.interpret(ctx, "close wooden box")
        assert printed == ["Container is already closed."]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_tell_contents_empty(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer):
        box = setup_container(t_wizard)
        box.tell_contents()
        assert printed == ["It is empty."]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_tell_contents_with_items(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer):
        system = lookup(1)
        box = setup_container(t_wizard)
        tobacco = create("tobacco", parents=[system.thing], location=box)
        box.tell_contents()
        assert printed == ["Contents:", f"  {tobacco.title()}"]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_put_in_open_container(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer) as ctx:
        system = lookup(1)
        box = setup_container(t_wizard)
        tobacco = create("tobacco", parents=[system.thing], location=t_wizard.location)
        parse.interpret(ctx, "open wooden box")
        printed.clear()
        parse.interpret(ctx, "put tobacco in wooden box")
        tobacco.refresh_from_db()
        assert tobacco.location == box
        assert printed == ["You placed tobacco in wooden box"]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_put_in_closed_container(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer) as ctx:
        system = lookup(1)
        box = setup_container(t_wizard)
        tobacco = create("tobacco", parents=[system.thing], location=t_wizard.location)
        parse.interpret(ctx, "put tobacco in wooden box")
        tobacco.refresh_from_db()
        assert tobacco.location != box
        assert f"{box.title()} is closed." in printed


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_take_from_open_container(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer) as ctx:
        system = lookup(1)
        lab = t_wizard.location
        box = setup_container(t_wizard)
        tobacco = create("tobacco", parents=[system.thing], location=box)
        parse.interpret(ctx, "open wooden box")
        printed.clear()
        parse.interpret(ctx, "take tobacco from wooden box")
        tobacco.refresh_from_db()
        assert tobacco.location == t_wizard
        assert printed == ["You took tobacco from wooden box"]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_take_from_closed_container(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer) as ctx:
        system = lookup(1)
        box = setup_container(t_wizard)
        tobacco = create("tobacco", parents=[system.thing], location=box)
        parse.interpret(ctx, "take tobacco from wooden box")
        tobacco.refresh_from_db()
        assert tobacco.location == box
        assert printed == [f"{box.title()} is closed."]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_take_item_not_in_container(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer) as ctx:
        system = lookup(1)
        box = setup_container(t_wizard)
        tobacco = create("tobacco", parents=[system.thing], location=t_wizard.location)
        parse.interpret(ctx, "open wooden box")
        printed.clear()
        parse.interpret(ctx, "take tobacco from wooden box")
        assert printed == ["tobacco is not in wooden box."]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_is_openable_by_no_key(t_init: Object, t_wizard: Object):
    """Container with no lock key is always openable."""
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer):
        box = setup_container(t_wizard)
        assert box.is_openable_by(t_wizard)


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_lock_for_open(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer) as ctx:
        system = lookup(1)
        box = setup_container(t_wizard)
        key = create("brass key", parents=[system.thing], location=t_wizard.location)
        parse.interpret(ctx, "lock_for_open wooden box with brass key")
        assert box.get_property("open_key") == key


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_unlock_for_open(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer) as ctx:
        system = lookup(1)
        box = setup_container(t_wizard)
        key = create("brass key", parents=[system.thing], location=t_wizard.location)
        parse.interpret(ctx, "lock_for_open wooden box with brass key")
        parse.interpret(ctx, "unlock_for_open wooden box")
        assert box.get_property("open_key") is None


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_open_locked_container_without_key(t_init: Object, t_wizard: Object):
    """Opening a locked container when the player does not hold the key is silently blocked."""
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer) as ctx:
        system = lookup(1)
        box = setup_container(t_wizard)
        key = create("brass key", parents=[system.thing], location=t_wizard.location)
        box.set_property("open_key", key.id)
        parse.interpret(ctx, "open wooden box")
        box.refresh_from_db()
    assert not box.is_open()
    assert printed == []


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_open_locked_container_with_key(t_init: Object, t_wizard: Object):
    """Opening a locked container while holding the required key succeeds."""
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer) as ctx:
        system = lookup(1)
        box = setup_container(t_wizard)
        key = create("brass key", parents=[system.thing], location=t_wizard)
        box.set_property("open_key", key.id)
        parse.interpret(ctx, "open wooden box")
        box.refresh_from_db()
    assert box.is_open()
    assert printed == ["You open the container."]



import pytest

from moo.core.models.object import Object

from .. import code


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_eval_simple_command(t_init: Object, t_wizard: Object):
    def _writer(msg):
        raise RuntimeError("print was called unexpectedly")

    with code.ContextManager(t_wizard, _writer):
        writer = code.ContextManager.get("writer")
        globals = code.get_default_globals()  # pylint: disable=redefined-builtin
        globals.update(code.get_restricted_environment("__main__", writer))
        result = code.do_eval("dir()", {}, globals)
        assert result == []


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_trivial_printing(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer):
        writer = code.ContextManager.get("writer")
        globals = code.get_default_globals()  # pylint: disable=redefined-builtin
        globals.update(code.get_restricted_environment("__main__", writer))
        result = code.do_eval("print('test')", {}, globals)
        assert result is None
        assert printed == ["test"]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_printing_imported_caller(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer):
        writer = code.ContextManager.get("writer")
        globals = code.get_default_globals()  # pylint: disable=redefined-builtin
        globals.update(code.get_restricted_environment("__main__", writer))
        src = "from moo.core import context\nprint(context.caller)"
        code.r_exec(src, {}, globals)
        assert printed == [t_wizard]

@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_caller_stack(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer):
        from moo.core import context, lookup, create

        player = lookup("Player")

        # Create 3 new objects to serve as verb owners
        p3 = create("TestPlayer3")
        p4 = create("TestPlayer4")
        p5 = create("TestPlayer5")

        # Build a chain of 5 verbs on the Player class, each owned by a different player.
        # add_verb() sets owner = ContextManager.get("caller") = t_wizard, so we
        # update the owner afterward. Permissions (allow "everyone" "execute") are set
        # automatically by apply_default_permissions during Verb.save() on creation.

        v1 = player.add_verb("test-caller-chain-1", code="""
from moo.core import context
this.invoke_verb("test-caller-chain-2")
""")
        v1.owner = t_wizard
        v1.save()

        v2 = player.add_verb("test-caller-chain-2", code="""
from moo.core import context
this.invoke_verb("test-caller-chain-3")
""")
        v2.owner = player
        v2.save()

        v3 = player.add_verb("test-caller-chain-3", code="""
from moo.core import context
this.invoke_verb("test-caller-chain-4")
""")
        v3.owner = p3
        v3.save()

        v4 = player.add_verb("test-caller-chain-4", code="""
from moo.core import context
this.invoke_verb("test-caller-chain-5")
""")
        v4.owner = p4
        v4.save()

        v5 = player.add_verb("test-caller-chain-5", code="""
from moo.core import context
for frame in context.caller_stack:
    print(frame)
""")
        v5.owner = p5
        v5.save()

        # Invoke the chain
        player.invoke_verb("test-caller-chain-1")

        # Stack is fully unwound after all verbs return
        assert context.caller_stack == []

        # The 5 frames were captured in order by verb 5
        assert printed == [
            dict(caller=t_wizard, origin=player, player=t_wizard, previous_caller=t_wizard, this=player, verb_name="test-caller-chain-1"),
            dict(caller=player, origin=player, player=t_wizard, previous_caller=t_wizard, this=player, verb_name="test-caller-chain-2"),
            dict(caller=p3, origin=player, player=t_wizard, previous_caller=player, this=player, verb_name="test-caller-chain-3"),
            dict(caller=p4, origin=player, player=t_wizard, previous_caller=p3, this=player, verb_name="test-caller-chain-4"),
            dict(caller=p5, origin=player, player=t_wizard, previous_caller=p4, this=player, verb_name="test-caller-chain-5"),
        ]



#!moo verb @eject --on $player --dspec any --ispec from:any

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is a player command used to remove objects from rooms or the player's inventory. The indirect object is used
to specify what the direct object is to be ejected from.

If neither of the special cases match, the direct and indirect object are matched on, to give an object reference to
use. Suitable error messages are printed if a match is not found. If matches are found, then a sequence of message
printing is started. The indirect object's `victim_ejection_msg` verb is invoked, and the returned result printed to
the victim. The string returned by the indirect object's `ejection_msg` is printed to the player doing the ejecting.
The result returned by the indirect object's `oejection_msg` verb is printed to everyone else in the room.

Finally, the indirect object's `eject` verb is called to remove the victim.
"""

from moo.core import context

parser = context.parser
player = context.player
if parser.has_pobj("from"):
    container = parser.get_pobj("from")
    victims = container.find(parser.get_dobj_str())
    if not victims:
        print(f"Couldn't find {parser.get_dobj_str()} in {container.name}.")
    else:
        victim = victims.first()
        victim.tell(container.victim_ejection_msg())
        print(container.ejection_msg())
        container.announce(container.oejection_msg(), exclude=[player, victim])
        container.eject(victim)
else:
    victim = parser.get_dobj()
    victim.tell(victim.location.victim_ejection_msg())



#!moo verb close --on $container --dspec this

# pylint: disable=return-outside-function,undefined-variable

"""
This is the opposite of open. If the pipe is already close, an error message is printed.
"""

from moo.core import context

if not this.is_open():
    print("Container is already closed.")
elif this.is_openable_by(context.player):
    this.set_opened(False)
    print("You close the container.")



# Generated by Django 5.0 on 2024-02-11 15:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0006_remove_object_observers_delete_observation"),
    ]

    operations = [
        migrations.AddField(
            model_name="property",
            name="inherited",
            field=models.BooleanField(default=False),
        ),
    ]



#!moo verb huh2 --on $room

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is called by the `huh` verb to provide default handling of unrecognized commands given to players
in this room. You can override this verb to provide custom handling of such commands. If you wish to
fall back to the default behavior, you can use `passthrough()` to call the parent room's `huh2` verb.
"""

from moo.core import context

context.player.tell("Huh? I don't understand that command.")



import logging

from moo import bootstrap
from moo.core import code, create, lookup
from moo.core.models import Player, User

log = logging.getLogger(__name__)

repo = bootstrap.initialize_dataset("test")
wizard = lookup("Wizard")
user = User.objects.create(username="wizard")
Player.objects.create(user=user, avatar=wizard, wizard=True)
containers = lookup("container class")

with code.ContextManager(wizard, log.info):
    bag = create("bag of holding", parents=[containers], location=wizard)
    hammer = create("wizard hammer", location=bag)
    book = create("class book", parents=[containers], location=bag)
    players = create("player class", location=book)
    guests = create("guest class", location=book)
    guests.parents.add(players)
    authors = create("author class", location=book)
    authors.parents.add(players)
    programmers = create("programmer class", location=book)
    programmers.parents.add(authors)
    wizards = create("wizard class", location=book)
    wizards.parents.add(programmers)

    wizard.parents.add(wizards)

    rooms = create("room class", parents=[containers], location=book)
    rooms.set_property("description", "There's not much to see here.")

    lab = create("The Laboratory")
    lab.parents.add(rooms)
    lab.set_property(
        "description",
        """A cavernous laboratory filled with gadgetry of every kind,
    this seems like a dumping ground for every piece of dusty forgotten
    equipment a mad scientist might require.""",
    )

    wizard.location = lab
    wizard.save()

    new_player = create(name="Player", unique_name=True, location=lab)
    p = Player.objects.create()
    p.avatar = new_player
    p.save()

    new_player.parents.add(players, containers)
    new_player.owner = new_player
    new_player.save()

    bootstrap.load_verbs(repo, "moo.bootstrap.test_verbs")



# -*- coding: utf-8 -*-
"""
A game server for hosting text-based online MOO-like games.
"""
# This will make sure the app is always imported when
# Django starts so that shared_task will use this app.
from .celery import app as celery_app

__all__ = ("celery_app", "get_version")

__version__ = "0.65.4"


def get_version():
    """
    Get the current version of DjangoMOO.
    """
    return __version__



#!moo verb take get --on $thing --dspec this

# pylint: disable=return-outside-function,undefined-variable

"""
One or the other of these verbs is invoked when a player tries to take possession of an object i.e., pick it up. The
code involved is fairly straightforward. It checks to see if the player already has the object, and prints a suitable
message if this is the case. If not, then the `moveto` verb on the object is invoked. If this results in the object
moving into the player's inventory, then the take_succeeded messages defined on the object are printed. If the `moveto`
failed, then the `*take_failed` messages for the object are printed.

This scheme allows you to add extra conditions to restrict whether a player can take an object or not. For example, you
may place a notion of strength onto a player, and add weight to objects. If an object is too heavy for a player to lift,
then the object cannot be taken by the player. This sort of condition should be added to the `take` verb for the object
"""

from moo.core import context

title = this.title()
if this.location == context.player:
    print("You already have {} in your inventory.".format(title))
elif this.moveto(context.player):
    print(this.take_succeeded_msg(title).format(actor="You", subject=this))
    if msg := this.otake_succeeded_msg(title):
        this.location.announce(msg.format(actor=context.player, subject=this))
else:
    print(this.take_failed_msg(title).format(actor="You", subject=this))
    if msg := this.otake_failed_msg(title):
        this.location.announce(msg.format(actor=context.player, subject=this))



#!moo verb @sethome --on $player

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is a player command used to set the player's home to be the his/her current location. You can only set your
home to a room if the room will accept you, determined by checking the room's `accept` verb. This allows builders to
restrict which rooms may be used by players as their homes. If the room does not allow the player to set it as his/her
home, a suitable message is printed to inform the user of this fact. Otherwise, the player's `home` property is set to
the player's current location.
"""

from moo.core import context

player = context.player
if not player.location.accept(player):
    print("You cannot set your home to this location. It is not allowing you to enter.")
    return
player.set_property("home", player.location)
print("Your home has been set to your current location.")



# Generated by Django 5.1.4 on 2025-05-11 19:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0015_remove_verb_ability_remove_verb_method'),
    ]

    operations = [
        migrations.AddField(
            model_name='prepositionspecifier',
            name='preposition_specifier',
            field=models.CharField(choices=[('any', 'any'), ('none', 'none')], db_index=True, default='none', max_length=255),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='prepositionspecifier',
            name='specifier',
            field=models.CharField(blank=True, choices=[('this', 'this'), ('any', 'any'), ('none', 'none')], db_index=True, max_length=255, null=True),
        ),
    ]



#!moo verb @opacity opacity --on $container --dspec this --ispec is:any

# pylint: disable=return-outside-function,undefined-variable

"""
Opacity determines when/if you can look at the contents of the container. There are three levels of opacity:

    0 - transparent
    1 - opaque
    2 - black hole

When the opacity is set to 0, you can see the contents when the pipe is open or closed. When the opacity is set to 1,
you can only see the contents when the container is open. If opacity is set to 2, you can never see the contents when
looking at the container.

The syntax for `@opacity' is:

    @opacity container is #

where '#' is either 0, 1 or 2.
"""

opacity = args[0] if args else context.parser.get_pobj_str("is")
if opacity not in ("0", "1", "2"):
    raise ValueError("Opacity must be 0, 1 or 2.")
this.set_opaque(int(opacity))



#!moo verb throw drop --on $thing --dspec this

# pylint: disable=return-outside-function,undefined-variable

"""
One or the other of these verbs is invoked when a player tries to place an object in the room s/he is currently
located, i.e., when the player tries to drop the object. Again, the code is fairly straightforward. If the object is
not located on the player, then a suitable message is sent to the player telling him/her to check his/her pockets. If
the player does possess the object, and the current location `accept` verb indicates that the room will allow the
object to be dropped, the object `moveto` verb is invoked to move the object from the player's inventory to the
contents list of the player's location. Suitable messages are printed to inform the player that the action succeeded,
and to tell other people in the room that the object has just been dropped.
"""

from moo.core import context

title = this.title()
if this.location != context.player:
    print("You check your pockets, but can't find {}.".format(title))
elif this.location.accept(this):
    this.moveto(context.player.location)
    print(this.drop_succeeded_msg(title).format(actor="You", subject=this))
    if msg := this.odrop_succeeded_msg(title):
        this.location.announce(msg.format(actor=context.player, subject=this))
else:
    print(this.drop_failed_msg(title).format(actor="You", subject=this))
    if msg := this.odrop_failed_msg(title):
        this.location.announce(msg.format(actor=context.player, subject=this))



#!moo verb @par*anoid --on $player --dspec any

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is used to set the player's paranoid level, as stored in the `paranoid` property. Three different levels are
available:

    `0`: The normal case, where no paranoia is applied to any messages sent to the player.
    `1`: In this case, the anti-spoofer is enabled, and the value of the `lines` property on the player is set to 20.
       This determines how many messages are stored on the player, for checking with the `@check' command.
    `2`: In this case, every message sent to the player is prefixed with the name and object number of the sender. This
       is the immediate mode of the anti-spoofer mechanism.
"""

from moo.core import context

level = context.parser.get_dobj_str()
if level not in ("0", "1", "2"):
    print("Paranoid level must be 0, 1, or 2.")
else:
    context.player.paranoid = int(level)



#!moo verb test-nested-verbs-method --on "player class" --dspec this

# pylint: disable=return-outside-function,undefined-variable

from moo.core import context

counter = 1
if len(args):  # pylint: disable=undefined-variable
    counter = args[0] + 1  # pylint: disable=undefined-variable

print(counter)

if counter < 10:
    context.caller.invoke_verb("test-nested-verbs-method", counter)



import pytest

from moo.core import code, lookup, parse
from moo.core.models import Object


# --- gag_p ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_gag_p_empty(t_init: Object, t_wizard: Object):
    """gag_p() returns False when the gag list is empty."""
    with code.ContextManager(t_wizard, lambda _: None):
        player_obj = lookup("Player")
        assert player_obj.gag_p() is False


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_gag_p_player_gagged(t_init: Object, t_wizard: Object):
    """gag_p() returns True when context.player is in this.gaglist."""
    with code.ContextManager(t_wizard, lambda _: None):
        player_obj = lookup("Player")
        player_obj.gaglist = [t_wizard]
        assert player_obj.gag_p() is True


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_gag_p_player_not_gagged(t_init: Object, t_wizard: Object):
    """gag_p() returns False when context.player is not in this.gaglist."""
    with code.ContextManager(t_wizard, lambda _: None):
        player_obj = lookup("Player")
        t_wizard.gaglist = [t_wizard]
        assert player_obj.gag_p() is False


# --- @gag ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_gag_player(t_init: Object, t_wizard: Object):
    """@gag <player> adds the player to the caller's gaglist."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        player_obj = lookup("Player")
        parse.interpret(ctx, "@gag Player")
        t_wizard.refresh_from_db()
    assert player_obj in t_wizard.gaglist
    assert "Gag list updated." in printed


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_gag_object(t_init: Object, t_wizard: Object, setup_item):
    """@gag <object> adds a non-player object to the caller's object_gaglist."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        widget = setup_item(t_wizard.location, "widget")
        parse.interpret(ctx, "@gag widget")
        t_wizard.refresh_from_db()
    assert widget in t_wizard.object_gaglist
    assert "Gag list updated." in printed


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_gag_already_gagged_player(t_init: Object, t_wizard: Object):
    """@gag on an already-gagged player prints a duplicate message."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        player_obj = lookup("Player")
        t_wizard.gaglist = [player_obj]
        parse.interpret(ctx, "@gag Player")
    assert "You are already gagging Player." in printed
    assert "No changes made to gag list." in printed


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_gag_already_gagged_object(t_init: Object, t_wizard: Object, setup_item):
    """@gag on an already-gagged object prints a duplicate message."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        widget = setup_item(t_wizard.location, "widget")
        t_wizard.object_gaglist = [widget]
        parse.interpret(ctx, "@gag widget")
    assert "You are already gagging widget." in printed
    assert "No changes made to gag list." in printed


# --- @ungag ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_ungag_player(t_init: Object, t_wizard: Object):
    """@ungag removes a player from the gag list."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        player_obj = lookup("Player")
        t_wizard.gaglist = [player_obj]
        parse.interpret(ctx, "@ungag Player")
        t_wizard.refresh_from_db()
    assert player_obj not in t_wizard.gaglist
    assert any("no longer gagging" in line for line in printed)


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_ungag_everyone(t_init: Object, t_wizard: Object, setup_item):
    """@ungag everyone clears both gag lists."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        player_obj = lookup("Player")
        widget = setup_item(t_wizard.location, "widget")
        t_wizard.gaglist = [player_obj]
        t_wizard.object_gaglist = [widget]
        parse.interpret(ctx, "@ungag everyone")
        t_wizard.refresh_from_db()
    assert t_wizard.gaglist == []
    assert t_wizard.object_gaglist == []
    assert "Gag lists cleared." in printed


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_ungag_not_gagged(t_init: Object, t_wizard: Object):
    """@ungag on a player not in any list prints 'You are not gagging'."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        parse.interpret(ctx, "@ungag Player")
    assert any("not gagging" in line for line in printed)


# --- @listgag ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_listgag_empty(t_init: Object, t_wizard: Object):
    """listgag with empty lists prints 'None' for each section."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        t_wizard.listgag()
    assert "Gagged players:" in printed
    assert "Gagged objects:" in printed
    assert printed.count("  None") == 2


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_listgag_with_player(t_init: Object, t_wizard: Object):
    """listgag shows a gagged player's name when the list is non-empty."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        player_obj = lookup("Player")
        t_wizard.gaglist = [player_obj]
        t_wizard.listgag()
    assert any("Player" in line for line in printed)


# --- tell ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_tell_normal(t_init: Object, t_wizard: Object):
    """tell() routes the message to the player via write(), emitting a ConnectionError warning."""
    with code.ContextManager(t_wizard, lambda _: None):
        with pytest.warns(RuntimeWarning, match="Hello, world!"):
            t_wizard.tell("Hello, world!")


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_tell_gagged_suppressed(t_init: Object, t_wizard: Object):
    """tell() suppresses the message when context.player is in this.gaglist."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        player_obj = lookup("Player")
        player_obj.gaglist = [t_wizard]
        player_obj.tell("You should not see this")
    assert printed == []


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_tell_paranoid_1_stores_responsible(t_init: Object, t_wizard: Object):
    """tell() with paranoid=1 appends callers and message to responsible."""
    with code.ContextManager(t_wizard, lambda _: None):
        player_obj = lookup("Player")
        player_obj.paranoid = 1
        with pytest.warns(RuntimeWarning, match="track me"):
            player_obj.tell("track me")
        player_obj.refresh_from_db()
    assert len(player_obj.responsible) > 0


# --- @paranoid ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_paranoid_sets_level(t_init: Object, t_wizard: Object):
    """@paranoid <level> sets player.paranoid to the given integer level."""
    with code.ContextManager(t_wizard, lambda _: None) as ctx:
        parse.interpret(ctx, "@paranoid 1")
        t_wizard.refresh_from_db()
    assert t_wizard.paranoid == 1


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_paranoid_invalid_level(t_init: Object, t_wizard: Object):
    """@paranoid with an out-of-range level prints an error message."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        parse.interpret(ctx, "@paranoid 5")
    assert "Paranoid level must be 0, 1, or 2." in printed


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_paranoid_level_2(t_init: Object, t_wizard: Object):
    """@paranoid 2 sets the paranoid property to 2."""
    with code.ContextManager(t_wizard, lambda _: None) as ctx:
        parse.interpret(ctx, "@paranoid 2")
        t_wizard.refresh_from_db()
    assert t_wizard.paranoid == 2


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_paranoid_reset_to_0(t_init: Object, t_wizard: Object):
    """@paranoid 0 resets the paranoid property to 0 (disabling tracking)."""
    with code.ContextManager(t_wizard, lambda _: None) as ctx:
        parse.interpret(ctx, "@paranoid 1")
        parse.interpret(ctx, "@paranoid 0")
        t_wizard.refresh_from_db()
    assert t_wizard.paranoid == 0



# -*- coding: utf-8 -*-
"""
Verb model
"""

import logging
import warnings

from django.conf import settings
from django.core import validators
from django.db import models

from moo import bootstrap
from .. import utils, lookup
from ..code import interpret, ContextManager
from .acl import AccessibleMixin

log = logging.getLogger(__name__)


class Preposition(models.Model):
    pass


class PrepositionName(models.Model):
    name = models.CharField(max_length=255)
    preposition = models.ForeignKey(Preposition, related_name="names", on_delete=models.CASCADE)


class PrepositionSpecifier(models.Model):
    preposition = models.ForeignKey(Preposition, related_name="+", on_delete=models.SET_NULL, blank=True, null=True)
    preposition_specifier = models.CharField(
        max_length=255,
        choices=settings.PREPOSITION_SPECIFIER_CHOICES,
        db_index=True,
    )
    specifier = models.CharField(
        max_length=255, choices=settings.OBJECT_SPECIFIER_CHOICES, db_index=True, default="none"
    )


class Verb(models.Model, AccessibleMixin):
    #: The Python code for this Verb
    code = models.TextField(blank=True, null=True)
    #: Optional Git repo this code is from
    repo = models.ForeignKey("Repository", related_name="+", blank=True, null=True, on_delete=models.SET_NULL)
    #: Optional name of the code file within the repo
    filename = models.CharField(max_length=255, blank=True, null=True)
    #: Optional Git ref of the code file within the repo
    ref = models.CharField(max_length=255, blank=True, null=True)
    #: The owner of this Verb. Changes require `entrust` permission.
    owner = models.ForeignKey("Object", related_name="+", blank=True, null=True, on_delete=models.SET_NULL)
    #: The object on which this Verb is defined
    origin = models.ForeignKey("Object", related_name="verbs", on_delete=models.CASCADE)
    #: If the Verb can be called with a direct obect
    direct_object = models.CharField(
        max_length=255, choices=settings.OBJECT_SPECIFIER_CHOICES, db_index=True, default="none", db_default="none"
    )
    #: If the Verb can be called with an indirect obect
    indirect_objects = models.ManyToManyField(PrepositionSpecifier, related_name="+", blank=True)

    do_not_call_in_templates = True
    invoked_object = None
    invoked_name = None

    def __str__(self):
        return "%s {#%s on %s}" % (self.annotated(), self.id, self.origin)

    @property
    def kind(self):
        return "verb"

    @property
    def is_ability(self):
        return self.direct_object == "this" and self.indirect_objects is None

    @property
    def is_method(self):
        return self.direct_object is None and self.indirect_objects is not None

    def annotated(self):
        ability_decoration = ["", "@"][int(self.is_ability)]
        method_decoration = ["", "()"][int(self.is_method)]
        verb_name = self.name()
        return "".join([ability_decoration, verb_name, method_decoration])

    def name(self):
        names = self.names.all()
        if not names:
            return "(untitled)"
        return names[0].name

    def save(self, *args, **kwargs):
        needs_default_permissions = self.pk is None
        super().save(*args, **kwargs)
        if not needs_default_permissions:
            return
        utils.apply_default_permissions(self)

    def reload(self):
        if self.repo is None or self.filename is None:
            raise RuntimeError("Cannot reload a verb without an associated repo and filename.")
        import pathlib
        bootstrap.load_verb_source(pathlib.Path(self.filename), lookup(1), self.repo, replace=True)

    def is_bound(self):
        return hasattr(self, "invoked_object") and hasattr(self, "invoked_name")

    def passthrough(self, *args, **kwargs):
        """
        Invoke this verb on the parent objects, if they exist.

        Often, it is useful for a child object to define a verb that augments the behavior of a verb on its parent
        object. For example, in the LambdaCore database, the root object (which is an ancestor of every other object)
        defines a verb called `description' that simply returns the value of this.description; this verb is used by
        the implementation of the look command. In many cases, a programmer would like the description of some object
        to include some non-constant part; for example, a sentence about whether or not the object was `awake' or
        `sleeping'. This sentence should be added onto the end of the normal description. The programmer would like to
        have a means of calling the normal description verb and then appending the sentence onto the end of that
        description. The function `passthrough()` is for exactly such situations.

        passthrough calls the verb with the same name as the current verb but as defined on the parent of the object
        that defines the current verb. The arguments given to passthrough are the ones given to the called verb and the
        returned value of the called verb is returned from the call to passthrough. The initial value of `this` in the
        called verb is the same as in the calling verb.

        Thus, in the example above, the child-object's description verb might have the following implementation:

            return passthrough() + "  It is " + (this.awake ? "awake." | "sleeping.")

        That is, it calls its parent's description verb and then appends to the result a sentence whose content is
        computed based on the value of a property on the object.

        In almost all cases, you will want to call `passthrough()' with the same arguments as were given to the current
        verb. This is easy to write in Python; just call passthrough(*args).
        """
        if not self.is_bound():
            raise RuntimeError("Cannot use passthrough on an unbound verb.")

        # the origin is where the verb is defined, not where it was found
        parents = self.origin.parents.all()
        for parent in parents:
            if parent.has_verb(self.invoked_name):
                verb = parent.get_verb(self.invoked_name)
                verb.invoked_object = self.invoked_object
                verb.invoked_name = self.invoked_name
                assert verb != self, "Infinite passthrough loop detected."
                return verb(*args, **kwargs)
        warnings.warn(
            "Passthrough ignored: no parent has verb %s" % self.invoked_name,
            RuntimeWarning,
        )

    def __call__(self, *args, **kwargs):
        this = None
        name = "__main__"
        if self.is_bound():
            this = self.invoked_object
            name = self.invoked_name
        else:
            this = self.origin
            name = self.name()
        if self.filename is not None:
            kwargs['filename'] = self.filename
        if this is None:
            raise RuntimeError(f"Cannot call {self} without a non-null 'this' context.")
        _system_key = "__system_object__"
        _cache = ContextManager.get_perm_cache()
        if _cache is not None and _system_key in _cache:
            system = _cache[_system_key]
        else:
            system = lookup(1)
            if _cache is not None:
                _cache[_system_key] = system
        active = ContextManager.is_active()
        if active:
            ContextManager.override_caller(self.owner, this=this, verb_name=name, origin=self.origin, player=ContextManager.get("player"))
        try:
            result = interpret(self.code, name, this, self.passthrough, system, *args, **kwargs)
        finally:
            if active:
                ContextManager.pop_caller()
        return result


class VerbName(models.Model):
    verb = models.ForeignKey(Verb, related_name="names", on_delete=models.CASCADE)
    name = models.CharField(max_length=255, db_index=True)

    class Meta:
        constraints = [models.UniqueConstraint("verb", "name", name="unique_verb_name")]

    def __str__(self):
        return "%s {#%s on %s}" % (self.name, self.verb.id, self.verb.origin)


# TODO: add support for additional URL types and connection details
class URLField(models.CharField):
    default_validators = [validators.URLValidator(schemes=["https"])]


class Repository(models.Model):
    slug = models.SlugField()
    url = URLField(max_length=255)
    prefix = models.CharField(max_length=255)
